__version__ = "2.8"   # ___PKG_VERSION___: DO NOT MODIFY the version here. Update it via version_manager.py!
__engine_version__ = "1.0.8"   # ___PKG_VERSION___: DO NOT MODIFY the version here. Update it via version_manager.py!
__engine_release_tag__ = "2.7"   # ___PKG_VERSION___: DO NOT MODIFY the version here. Update it via version_manager.py!
